from typing import TYPE_CHECKING, Any, ClassVar

from ixmp4 import db
from ixmp4.core.exceptions import OptimizationDataValidationError
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract

from .. import base, utils

if TYPE_CHECKING:
    from .. import IndexSet


class TableIndexsetAssociation(base.RootBaseModel):
    __tablename__ = "optimization_tableindexsetassociation"

    table__id: types.TableId
    table: types.Mapped["Table"] = db.relationship(
        back_populates="_table_indexset_associations"
    )
    indexset__id: types.IndexSetId
    indexset: types.Mapped["IndexSet"] = db.relationship()

    column_name: types.String = db.Column(db.String(255), nullable=True)


class Table(base.BaseModel):
    __tablename__ = "optimization_table"

    # NOTE: These might be mixin-able, but would require some abstraction
    NotFound: ClassVar = abstract.Table.NotFound
    NotUnique: ClassVar = abstract.Table.NotUnique
    DataInvalid: ClassVar = OptimizationDataValidationError
    DeletionPrevented: ClassVar = abstract.Table.DeletionPrevented

    run__id: types.RunId
    data: types.JsonDict = db.Column(db.JsonType, nullable=False, default={})

    @db.validates("data")
    def validate_data(self, key: Any, data: dict[str, Any]) -> dict[str, Any]:
        if not bool(data):
            return data
        utils.validate_data(
            host=self,
            data=data,
            indexsets=self._indexsets,
            column_names=self.column_names,
            has_extra_columns=False,
        )
        return data

    _table_indexset_associations: types.Mapped[list[TableIndexsetAssociation]] = (
        db.relationship(
            back_populates="table",
            cascade="all, delete-orphan",
            order_by="TableIndexsetAssociation.id",
            passive_deletes=True,
        )
    )

    _indexsets: db.AssociationProxy[list["IndexSet"]] = db.association_proxy(
        "_table_indexset_associations", "indexset"
    )
    _column_names: db.AssociationProxy[list[str | None]] = db.association_proxy(
        "_table_indexset_associations", "column_name"
    )

    @property
    def indexset_names(self) -> list[str]:
        return [indexset.name for indexset in self._indexsets]

    @property
    def column_names(self) -> list[str] | None:
        names = [name for name in self._column_names if name]
        return names if bool(names) else None

    __table_args__ = (db.UniqueConstraint("name", "run__id"),)



from typing import TYPE_CHECKING, Any, Union

import pandas as pd

from . import base

if TYPE_CHECKING:
    from .equation import Equation
    from .indexset import IndexSet
    from .parameter import Parameter
    from .table import Table
    from .variable import Variable


def validate_data(
    host: base.BaseModel,
    data: dict[str, Any],
    indexsets: list["IndexSet"],
    column_names: list[str] | None = None,
    has_extra_columns: bool = True,
) -> None:
    data_frame = pd.DataFrame.from_dict(data)

    # Can't validate ("values","units") or ("levels", "marginals") when they are present
    number_columns = (
        len(data_frame.columns) - 2 if has_extra_columns else len(data_frame.columns)
    )
    columns = (
        column_names if column_names else [indexset.name for indexset in indexsets]
    )

    # TODO for all of the following, we might want to create unique exceptions
    # Could me make both more specific by specifiying missing/extra columns?
    if number_columns < len(indexsets):
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \n"
            f"Data is missing for some columns! \n Data: {data} \n "
            f"Columns: {columns}"
        )
    elif number_columns > len(indexsets):
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \n"
            f"Trying to add data to unknown columns! \n Data: {data} \n "
            f"Columns: {columns}"
        )

    # We could make this more specific maybe by pointing to the missing values
    if data_frame.isna().any(axis=None):
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \n"
            "The data is missing values, please make sure it "
            "does not contain None or NaN, either!"
        )

    limited_to_indexsets = {
        columns[i]: indexsets[i].data for i in range(len(indexsets))
    }

    # We can make this more specific e.g. highlighting all duplicate rows via
    # pd.DataFrame.duplicated(keep="False")
    if data_frame[limited_to_indexsets.keys()].value_counts().max() > 1:
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \nThe data contains duplicate rows!"
        )

    # Can we make this more specific? Iterating over columns; if any is False,
    # return its name or something?
    if (
        not data_frame[limited_to_indexsets.keys()]
        .isin(limited_to_indexsets)
        .all(axis=None)
    ):
        raise host.DataInvalid(
            f"While handling {host.__str__()}: \n"
            "The data contains values that are not allowed as per the IndexSets "
            "it is constrained to!"
        )


# NOTE | does not seem to work yet for stringified type hints
def _find_columns_linked_to_indexset(
    item: Union["Table", "Parameter", "Equation", "Variable"], name: str
) -> list[str]:
    """Determine columns in `item`.data that are linked to IndexSet `name`.

    Only works for `items` linked to IndexSet `name`.
    """
    if not item.column_names:
        # The item's indexset_names must be a unique list of names
        return [name]
    else:
        # If we have column_names, we must also have indexsets
        assert item.indexset_names

        # Handle possible duplicate values
        return [
            item.column_names[i]
            for i in range(len(item.column_names))
            if item.indexset_names[i] == name
        ]



from typing import ClassVar

from ixmp4 import db
from ixmp4.data import types
from ixmp4.data.abstract import optimization as abstract
from ixmp4.data.db.unit import Unit

from .. import base


class Scalar(base.BaseModel):
    __tablename__ = "optimization_scalar"

    NotFound: ClassVar = abstract.Scalar.NotFound
    NotUnique: ClassVar = abstract.Scalar.NotUnique
    DeletionPrevented: ClassVar = abstract.Scalar.DeletionPrevented

    run__id: types.RunId

    value: types.Float = db.Column(db.Float, nullable=True, unique=False)

    unit: types.Mapped[Unit] = db.relationship()
    unit__id: types.Mapped[int] = db.Column(
        db.Integer, db.ForeignKey("unit.id"), index=True
    )

    __table_args__ = (db.UniqueConstraint("name", "run__id"),)
